from .camera.camera import Camera
from .camera.camera_calibration import CameraCalibration
from .connections import *
from .connections.network_location import NetworkLocation
from .connections.network_tables_connection import NetworkTablesConnection
from .connections.serial_connection import SerialConnection
from .target_filters import *
from .directions.directing_functions import *
from .directions.director import Director
from .direction_modifiers.direction_modifier import DirectionModifier
from .direction_modifiers.stop_if_close_modifier import StopIfCloseModifier
from .image_utilities.display_image import display_image
from .exceptions import exceptions
from .image_filters.image_filters import *
from .image_utilities.naming_conventions import *
from .image_utilities.photo_array import photo_array
from .math.geometry import *
from .math.default_math_functions import *
from .math.contours import *
from .math.shape_fill_ratios import *
from .math.image import *
from .math.image import image_size
from .morphological_functions.morphological_functions import erosion, dilation
from .thresholds.canny_edge import CannyEdge
from .thresholds.color.built_in_colors import *
from .thresholds.color.color import Color
from .thresholds.color.multi_color import MultiColor
from .thresholds.threshold import Threshold
from .visions.ambient_vision import AmbientVision
from .visions.multi_vision import MultiVision
from .visions.vision import Vision

__version__ = '2021.1.0'



from typing import Tuple
import math
import cv2

from ..math.contours import contour_center
from ..math.geometry import distance_between_points
from ..math.image import distance_from_frame
from .contour_filter import contour_filter
from ..helpers.types import RangedNumber
from ..math import image


@contour_filter
def image_center_filter(contour_list, image_dimensions: Tuple[int, int] = (320, 240),
                        min_ratio: RangedNumber(0, 1) = 0.7, max_ratio: RangedNumber(0, 1) = math.inf):
    """
    Filters out contours that their center is not close enough to the center of the image

    :param contour_list: a list of contours to be filtered
    :param image_dimensions: the size of the image(width, height)
    :param min_ratio: the minimum ratio between the distance from the image center to the distance from the frame
    :param max_ratio: the maximum ratio between the distance from the image center to the distance from the frame
    :return: list of contours within the maximum distance from the image center
    """
    output = []
    ratio = []
    image_center = image.image_center(image_dimensions)
    for contour in contour_list:
        current_contour_center = contour_center(contour)
        distance_from_image_center = distance_between_points(current_contour_center, image_center)
        distance_from_image_frame = distance_from_frame(current_contour_center, image_dimensions)
        distance_ratio = distance_from_image_center / distance_from_image_frame
        if min_ratio <= distance_ratio <= max_ratio:
            output.append(contour)
            ratio.append(distance_ratio)
    return output, ratio


@contour_filter
def distance_filter(contour_list, point: Tuple[int, int], min_dist: float = 0, max_dist: float = 50):
    """
    Filters out contours that their center is not close enough
    to the given (x, y) point in the image

    :param contour_list: a list of contours to be filtered
    :param point: the point from which the contours should be filtered a tuple (or list) of 2 numbers.
    :param max_dist: the maximum distance from the point in pixels
    :param min_dist: the minimum distance from the point in pixels
    :return: the filtered contour list
    """
    output = []
    ratio = []
    for contour in contour_list:
        current_contour_center = contour_center(contour)
        distance_from_center = distance_between_points(current_contour_center, point)
        if max_dist >= distance_from_center >= min_dist:
            output.append(contour)
            ratio.append(distance_from_center)
    return output, ratio


@contour_filter
def absolute_distance_filter(contour_list, max_dist=50, min_dist=0, image_dimensions=(320, 240)):
    """
    Filters out contours that their center is not close enough
    to the center of the image

    :param contour_list: a list of contours to be filtered
    :param image_dimensions: the size of the image(width, height)
    :param max_dist: the maximum distance from the center in pixels
    :param min_dist: the minimum distance from the center in pixels
    :return:
    """
    output = []
    ratio = []
    image_center = (image_dimensions[0] / 2 - .5, image_dimensions[1] / 2 - .5)
    for contour in contour_list:
        current_contour_center = contour_center(contour)
        distance_from_center = distance_between_points(current_contour_center, image_center)
        if max_dist >= distance_from_center >= min_dist:
            output.append(contour)
            ratio.append(distance_from_center)
    return output, ratio


@contour_filter
def length_filter(contour_list, min_length=50, max_length=76800):
    """
    Receives a list of contours and removes ones that are not long enough
    Note: for "open" contours only!

    :param contour_list: list of contours (numpy array) to be filtered
    :param min_length: minimum length of a contour (in pixels)
    :param max_length: maximum length of a contour (in pixels)
    :return: list of filtered contours and list of the lengths
    """
    output = []
    ratio = []
    for contour in contour_list:
        perimeter = cv2.arcLength(contour, False)
        if min_length >= perimeter >= max_length:
            output.append(contour)
            ratio.append(perimeter)
    return output, ratio


@contour_filter
def area_filter(contour_list, min_area: float = 200, max_area: float = math.inf):
    """
    Filters contours that are not within the threshold of area (in pixels)

    :param max_area: maximum area of a contour (Inclusive) default is no limit (infinity)
    :param min_area: minimum area of a contour (Inclusive) set to 0 for no lower limit.
    :param contour_list: List of Contours to filter
    :return: the contour list filtered.
    """
    output_list = []
    ratio_list = []
    for contour in contour_list:
        contour_area = cv2.contourArea(contour)
        if min_area <= contour_area <= max_area:
            output_list.append(contour)
            ratio_list.append(contour_area)
    return output_list, ratio_list


@contour_filter
def percent_area_filter(contour_list, minimal_percent: RangedNumber(0, 1) = 0.02,
                        maximum_percent: RangedNumber(0, 1) = 1, image_dimensions: Tuple[int, int] = (320, 240)):
    """
    Filters out contours that are not in the specified ratio to the area of the image (1% -> 0.1)

    :param contour_list: list of contours to be filtered (numpy.ndarray)
    :param minimal_percent: the minimal ratio between the contour area and the image area
    :param maximum_percent: the maximum ratio between the contour area and the image area
    :param image_dimensions: The (width, height) of the image in pixels,
    """
    output, ratios = [], []
    output_append = output.append
    ratio_append = ratios.append
    image_size = image_dimensions[0] * image_dimensions[1]
    if image_size == 0:
        raise ValueError("Invalid image dimensions, Received (width, height): {}, {}".format(*image_dimensions))
    for contour in contour_list:
        contour_area = cv2.contourArea(contour)
        percent_area = contour_area / image_size
        if minimal_percent <= percent_area <= maximum_percent:
            ratio_append(percent_area)
            output_append(contour)
    return output, ratios


@contour_filter
def size_ratio_filter(contours, min_ratio: float = 2, max_ratio: float = math.inf, reverse_ratio=False):
    """
    Sorts out contours by the ratio between their width and height

    :param contours: the contours to be filtered
    :param min_ratio: the minimum ratio
    :param max_ratio: the maximum ratio default
    :param reverse_ratio: reversed the ratio to be height to width
    :return: the filtered list
    """
    output, ratios = [], []
    output_append = output.append
    ratios_append = ratios.append
    for contour in contours:
        _, _, width, height = cv2.boundingRect(contour)
        if 0 in (width, height):
            raise ValueError(
                "The width or height of one of the contours was 0,\n"
                " try using an area filter before this filter")
        size_ratio = height / width if reverse_ratio else width / height
        if min_ratio <= size_ratio <= max_ratio:
            output_append(contour)
            ratios_append(size_ratio)
    return output, ratios



from functools import reduce
import math
import types
import cv2
import numpy as np
from typing import List, Union, Tuple, Any

from ..helpers.vision_detector_arguments import arguments_to_detector
from ..thresholds.threshold import Threshold
from ..detectors.detector import Detector
from ..exceptions.exceptions import InvalidCustomFunctionError, CameraError
from ..camera.camera import Camera
from ..partials.filter_applier import filter_applier
from ..connections.connection import Connection
from ..directions.director import Director
from ..camera.camera_settings import CameraSettings
from ..connections.network_location import NetworkLocation
from ..directions.directing_functions import center_directions


class Vision:
    """
    Vision object represents a computer vision pipeline.
    The pipeline consists of 4 main stages:
         1. processing - 'apply_all_image_filters' which uses a list of image_filter functions
         2. detection - 'detect' which comes from Detector objects and detects objects (contours, bounding rectangles..)
         3. filtering  - 'apply_target_filters' which is uses filter_functions like contour_filters
         4. conversion & usage - 'direct' which comes from director objects

    Each functionality can be used to easily create a complex yet modular

    Additional capabilities and tuning options are:
        Image filters (Blurs, rotations, cropping),
        Morphological functions,
        Ovl color HSVCalibration,

        1. camera handling
        2. connection clean-up and sending

    Vision can also be used as a part of a more complex pipeline.

    MultiVision can contain multiple vision objects and switch between pipelines, allowing for very versatile logic
    that can fit multiple needs.

    Ambient Vision is another option for using 2 different Vision objects and alternate between the 2.
    """

    def __init__(self, detector: Detector = None, threshold: Threshold = None,
                 morphological_functions: List[types.FunctionType] = None,
                 target_filters: List[types.FunctionType] = None,
                 director: Director = None, width=320, height=240, connection: Connection = None,
                 camera: Union[int, str, Camera, cv2.VideoCapture, Any] = None,
                 camera_settings: CameraSettings = None, image_filters: List[types.FunctionType] = None,
                 ovl_camera: bool = False, haar_classifier: str = None):
        """
        :param detector: a Detector object responsible for detecting targets
        :param threshold: threshold object, creates the binary mask from a given image, this cannot be passed together
        with detector
        :param target_filters: the list of contour_filter functions that
                                remove contours that aren't the target(s)
        :param director: a functions that receives a list or a single contour and returns director
        :param width: the width (in pixels) of images taken with the camera
        :param height: the height (in pixels)
        :param connection: a connection object that passes the result to the connection target
        :param camera: a Camera object (cv2.VideoCapture, ovl.Camera) or source from which to open a camera
        :param camera_settings: Special camera settings like calibration or offset used for
                                image correction and various direction calculations.
        :param image_filters: a list of image altering functions that are applied on the image.
        :param ovl_camera: a boolean that makes the camera opened to be ovl.Camera instead of cv2.VideoCapture
        :param haar_classifier:
        """
        mutually_exclusive_arguments = {"threshold": (threshold, morphological_functions),
                                        "detector": (detector,),
                                        "haar_cascade": (haar_classifier,)}

        detector = arguments_to_detector(mutually_exclusive_arguments)
        self.detector = detector
        self.width = width
        self.height = height
        self.target_filters = target_filters or []
        self.director = director or Director(center_directions, failed_detection=9999, target_amount=1)
        self.connection = connection
        self.image_filters = image_filters or []
        self.camera = None
        self.camera_port = None
        self.camera_settings = camera_settings

        if isinstance(camera, (cv2.VideoCapture, Camera)):
            self.camera = camera
        elif camera is None:
            pass
        else:
            self.camera_setup(camera, width, height, ovl_camera=ovl_camera)

    def __repr__(self):
        return str(self)

    def __str__(self):
        filters = [filter_function.__name__ for filter_function in self.target_filters]
        return "Vision: \n Detector: {} \n Filters: {}".format(self.detector, filters)

    @property
    def target_amount(self):
        """
        The wanted amount of targets
        Determined by self.director
        (0 None or math.inf if there is no limit, 1 if 1 target is wanted etc.)

        """
        if self.director is None:
            return math.inf
        return self.director.target_amount

    def send(self, data: Any, *args, **kwargs) -> Any:
        """
        Sends data to the destination using self.connection

        :param data: The data to send to the Connection
        :param args: any other arguments for the send function in your connection
        :param kwargs: any other named arguments for the connection object
        :return: Depends on the connection object used, returns its result

        """
        return self.connection.send(*args, **kwargs, data=data)

    def send_to_location(self, data: Any, network_location: NetworkLocation, *args, **kwargs):
        """
        A function that sends data to a specific NetworkLocation

        :param data: the data to be sent
        :param network_location: information used to send the data to a specific 'location'
         in the network
        :return: Depends on the connection object

        """
        return self.connection.send_to_location(data, network_location, *args, **kwargs)

    def get_image(self) -> np.ndarray:
        """
        Gets an image from self.camera and applies image filters

        :return: the image, false if failed to get it

        """
        if self.camera is None:
            raise ValueError("No camera given, (Camera is None)")
        if not self.camera.isOpened():
            raise CameraError("Camera given is not open (Has it been closed or disconnected?)")
        output = self.camera.read()
        if len(output) == 2:
            ret, image = output
            return image if ret else False
        else:
            return output

    def apply_target_filter(self, filter_function, contours, verbose=False):
        """
        Applies a filter function on the contour list, this is used to remove contours
        that do not match desired features

        NOTE: Vision.detect is mainly used for full object detection and filtering,
        refer to it for common use of Vision

        :param filter_function: Filter functions are function with a contour list variable that apply some
         sort of filter on the contours, thus removing ones that don't fit the limit given by the filter.
         for example: straight_rectangle_filter removes contours that are not rectangles that are parallel
         to the frame of the picture
        :param contours: the contours on which the filter should be applied (list of numpy.ndarrays)
        :param verbose: if true_shape does not print anything
        :return: returns the output of the filter function.

        """
        if verbose:
            print('Before "{}": {}'.format(filter_function.__name__, len(contours)))
        filter_function_output = filter_function(contours)

        if isinstance(filter_function_output, tuple):
            if len(filter_function_output) == 2:
                filtered_contours, ratio = filter_function_output[0], filter_function_output[1]
            else:
                raise InvalidCustomFunctionError('Filter function must return between 1 and 2 lists.'
                                                 'Please refer to the Documentation: '
                                                 'https://github.com/1937Elysium/Ovl-Python')
        elif isinstance(filter_function_output, list):
            filtered_contours, ratio = filter_function_output, []
        else:
            raise TypeError('The contour list must be a list or tuple of 2 lists (contours and ratios)')
        return filtered_contours, ratio

    def apply_target_filters(self, targets: List[np.ndarray], verbose=False
                             ) -> Tuple[List[np.ndarray], List[float]]:
        """
        Applies all of the filters on a list of contours, one after the other.
        Applies the first filter and passes the output to the second filter,

        :param targets: List of targets (numpy arrays or bounding boxes) to
        :param verbose: prints out information about filtering process if true (useful for debugging)
        :return: a list of all of the ratios given by the filter function in order.

        """
        ratios = []
        for filter_func in self.target_filters:
            targets, ratio = self.apply_target_filter(filter_func, targets, verbose=verbose)
            ratios.append(ratio)
        if verbose:
            print("After all filters: {}".format(len(targets)))
        return targets, ratios

    def apply_image_filters(self, image: np.ndarray) -> np.ndarray:
        """
        Applies all given image filters to the given image
        This is used to apply various image filters on your image in a pipeline,
        like blurs, image cropping, contrasting, sharpening, rotations, translations etc.

        :param image: the image that the image filters should be applied on (numpy array)
        :return: the image with the filters applied
        """
        return reduce(filter_applier, self.image_filters, image)

    def get_directions(self, contours: List[np.ndarray], image: np.ndarray, sorter=None):
        """
        Calculates the directions, based on contours found in the given image

        :param contours: final contours after filtering
        :param image: the image from which to find the contours
        :param sorter: optional parameter, applies a sorter on the given contours
        :return: a string of the director (output of the director function),
                 length depends on the director function
        """
        return self.director.direct(contours, image, sorter=sorter)

    def camera_setup(self, source=0, image_width=None, image_height=None, ovl_camera=False):
        """
        Opens up the camera reference and sets a given width and height to all images taken

        :param image_width: the width of the images to be taken, 0 does not set a width
        :param image_height: the height of the images to be taken, 0 does not set a height
        :param source: the location from which to open the camera
         string for network connections int for local USB connections.
        :param ovl_camera: if the camera object should be ovl.Camera
        :return: the camera object, also sets self.camera to the object.
        """

        image_height = image_height or self.height
        image_width = image_width or self.width
        self.camera_port = source
        if ovl_camera:
            camera = Camera(source=source, image_width=image_width, image_height=image_height)
        else:
            camera = cv2.VideoCapture(source)
            if image_width != -1:
                camera.set(3, image_width)
            if image_height != -1:
                camera.set(4, image_height)

        if not camera.isOpened():
            raise CameraError("Camera did not open correctly! Camera source: {}".format(self.camera_port))
        self.camera = camera
        return camera

    def detect(self, image, verbose=False, *args, **kwargs):
        """
        This is the function that performs processing detection and filtering on a given image, essentially passing
        the image through the detection related part of the pipeline

        detect applies image filters, detects objects in the filtered images (using the passed/created detector object)
        and finally applies all of the target_filters on the image.

        args and kwargs are passed to the detect function (passed to the detect method of the detector)

        :param verbose: passes verbose to apply_target_filters, which prints out information about the target filtering.
        :param image: image in which the vision should detect an object
        :return: contours and the filtered image and the ratios if return_ratios is true

        """
        image = self.apply_image_filters(image)
        targets = self.detector.detect(image, verbose, *args, **kwargs)
        return self.apply_target_filters(targets, verbose)



from numpy import ndarray

from ..math import contours as contours_


def xy_normalized_directions(contours, image: ndarray):
    """
    Returns the direction the detect objects (contours) are compared to the center
    of the image, this is returned in normalized screen space -1 to 1 (-1 meaning the most left compared to the center,
    and 1 the most right compared to the center and 0 meaning perfectly centered

    xy_normalized_directions returns x and y center directions,

    For example:
    If the center of the targets are the center of the image (width / 2, height / 2)
    the returned values are 0 and 0 meaning the camera is centered on the target


    :param contours: the final contours found after filtering - your targets
    :param image: the image from which it was found
    :return: the normalized screen space x and y coordinates of your final targets
    """
    return contours_.calculate_normalized_screen_space(contours, image)


def center_directions(contours, image: ndarray):
    """
    Returns the average center of the contours
    or list of contours that are the final targets

    This is the default directions function since it
    doesnt calculate any directions, only finds the center

    :param contours: the final contours - your targets
    :param image: the image from which it was found
    """
    return contours_.contour_average_center(contours)


def target_amount_directions(contours, image: ndarray):
    """
    Counts the amount of successful object detections.

    :param contours: the final contours - your targets
    :param image: the image from which it was found
    :return: the amount of targets detected
    """
    return len(contours)



import numpy as np


def is_greyscale_image(image: np.ndarray) -> bool:
    """
    A function that checks if an image is a greyscale image (an image where pixels are a number - has 1 color channel)

    :param image: the image to be checked, a numpy array
    """
    return len(image.shape) == 2


is_image_grayscale = is_greyscale_image


def is_color_image(image: np.ndarray) -> bool:
    """
    A function that checks if an image is a color image (each pixel is not one value but multiple color components)

    :param image: the image to be checked, a numpy array
    """
    return len(image.shape) == 3



import cv2
from pathlib import Path

from ..helpers.remove_none_values import remove_none_values


def open_image(image_path, flags=None):
    """
    Opens an image file, has a more direct error detection to find out problems
    that occurred during image opening, has a cost in performance.
    
    :param image_path: the path of the image to open
    :param flags: there are multiple flags that can be used to alter how the image is loaded
     https://docs.opencv.org/4.3.0/d4/da8/group__imgcodecs.html#ga288b8b3da0892bd651fce07b3bbd3a56
    :return: the opened image or a iterator that opens the list of images given
    """
    arguments = {"flags": flags}

    if not Path(image_path).is_file():
        raise OSError("Given image at path {} does not exist or is not a file!".format(image_path))

    image = cv2.imread(image_path, **remove_none_values(arguments))
    if image is None:
        raise ValueError("Couldn't read the image file, Check that {}\n"
                         "exists, that it's a valid format and that you\n"
                         "have the proper permissions to read it.".format(image))
    return image



from enum import Enum
from typing import Any, Union

from .connection import Connection


class HTTPRequestTypes(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    HEAD = "HEAD"


class HTTPConnection(Connection):
    """
    A Connection that utilizes http

    HTTP is useful for communicating with HTTP Servers over the web, although slower than other more
    low level connection types
    """
    def __init__(self, session, default_url=None, headers=None, auth=None):
        import requests
        self.session = session or requests.Session()
        self.session.headers.update(headers)
        if auth:
            self.auth = auth
            self.session.auth(auth)
        self.url = default_url

    def send_request(self, data: Any, url: str, method: HTTPRequestTypes, **kwargs):
        """
        send_request is the function used to send a request using http protocol.
        This can be used to send request to servers running remotely in the internet or to a HTTP server in the LAN.


        Note: This is the inner function used by the Connection object to send and receive.
        The functions send, receive and interval_receive should be used instead!


        :param data: The data to be passed as the request's payload
        :param url: the url to send the request to
        :param method: the HTTP method to be used
        :param kwargs: any other parameters to requests.request
        :return: the response received for the request
        """
        import requests
        params = {"data": data}
        if "params" in kwargs.keys():
            params = {**kwargs["params"], "data": data}
        url = url or self.url
        response = requests.request(method.value, url, params=params, **kwargs)
        response.raise_for_status()
        return response

    def send(self, data: Any, url: str = None, method=HTTPRequestTypes.POST, **kwargs):
        """
        Sends an http requests, should be used to only send data and not receive any data
        as it represent the logical action of sending information

        :param data: the data to be send, passed
        :param url:
        :param method:
        :param kwargs:
        :return:
        """
        return self.send_request(data, url, method, **kwargs)

    def receive(self, data: Any = None, url: str = None, method: HTTPRequestTypes = HTTPRequestTypes.GET,
                interval: Union[float, bool] = False, **kwargs):
        """
        Receive for HTTPConnection is a bit different compared to other Connection in that it sends a request
        and returns a response which is the main usage, in addition the interval parameter allows for high
        interval sampling (for MultiVision for example) without sending many requests

        :param url: the url to send to, overrides self.url
        :param data: any data to pass with the requests
        :param method: determines what http method should the request use (GET, POST, PUT etc.)
        :param interval:
        :param kwargs: any other parameters passed to Request.request
        :return:
        """
        if interval:
            return self.interval_receive(data, url, method, **kwargs)
        else:
            self.send_request(data, url, method, **kwargs)

    def interval_receive(self, data: Any, url: str = None, method=HTTPRequestTypes.GET, **kwargs):
        pass



from typing import List, Union

import cv2
from numpy import vstack, hstack, ndarray

WINDOWS = []


def stitch_images(images):
    width, height = images[0].shape
    stitched_image = cv2.imread(images[0]) if type(images[0]) is str else images[0]
    for image in images[1:]:
        if type(image) is str:
            image = cv2.imread(image)
        image_shape = image.shape
        if width + image_shape[0] > height + image_shape[1]:
            stitched_image = vstack((stitched_image, image))
            width += image_shape[0]
        else:
            stitched_image = hstack((stitched_image, image))
            height += image_shape[1]
    return stitched_image


def show_image(image, window_name, delay):
    cv2.imshow(window_name, image)
    return cv2.waitKey(delay)


def display_image(image: Union[ndarray, str, List[Union[ndarray, str]]], window_name='image', display_loop=False,
                  resizable=False):
    """
    The function displays an image

    The function is based on cv2.imread()
    is able to open both paths and numpy arrays (already opened) images

    When using this function to display images in a loop, delay=0 will cause the loop to stop until
    a key is pressed.

    .. code-block:: python

        for frame in video_frames:
            ovl.display_image(frame, delay=1)
            # the video will play normally in one window

    If delay is 0:

    .. code-block:: python

        for frame in video_frames:
            ovl.display_image(frame)
            # the video will freeze until a key is pressed



    :param image: Represents an image path (string), an already open image in the for of a numpy array (ndarray)
                  or a list of images (strings and arrays are valid)
    :param window_name: Name of the Window that displays the images
    :param display_loop: if display is used in a loop should be true, else false
    :param resizable: A boolean that determines whether the image window is resizable or not
    :return: the key pressed while image was displayed
    """
    if isinstance(image, str):
        image = cv2.imread(image)
    elif isinstance(image, list):
        image = stitch_images(image)
    elif not isinstance(image, ndarray):
        raise TypeError("Invalid image given, image must be an image, a path to an image or a list of images")

    if resizable:
        cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)

    return show_image(image, window_name, display_loop * 1)



import typing

import cv2
import numpy as np

from .geometry import circle_area


def rectangle_fill_ratio_straight(contour, reverse_div=False):
    """
    Returns the ratio between the contour and the straight bounding rectangle.

    :param contour: The contour to be compared
    :param reverse_div: If the rectangle's area should be divided by the contour.
    :return:  the ratio, the width of the bounding rectangle, the height of the bounding rectangle
    :rtype: float
    """
    _, _, width, height = cv2.boundingRect(contour)
    bounding_rectangle_area = width * height
    contour_area = cv2.contourArea(contour)
    if reverse_div:
        return float(bounding_rectangle_area) / contour_area, width, height
    return float(contour_area) / bounding_rectangle_area, width, height


def triangle_fill_ratio(contour, triangle=None) -> float:
    """
    Returns the ratio between a given contour and the smallest enclosing rectangle

    :param contour: numpy array
    :param triangle: convex hull for the minEnclosingTriangle function, can be ignored
    :return: returns the ratio between the contour and the bounding triangle
    :rtype: float
    """
    bounding_triangle_area, triangle = cv2.minEnclosingTriangle(contour, triangle)
    contour_area = cv2.contourArea(contour)
    return contour_area / float(bounding_triangle_area)


def rotating_rectangle_fill_ratio(contour: np.ndarray) -> typing.Tuple[float, float, float]:
    """
    Returns the ratio between the contour and the smallest bounding rectangle.

    :param contour: The contour to be compared
    :return: the ratio, the width of the bounding rectangle, the height of the bounding rectangle
    :rtype: float, float, float
    """
    rotated = cv2.minAreaRect(contour)
    bounding_rectangle_area = rotated[1][0] * rotated[1][1]
    contour_area = cv2.contourArea(contour)
    return float(contour_area) / bounding_rectangle_area, rotated[1][0], rotated[1][1]


def circle_fill_ratio(contour: np.ndarray) -> typing.Tuple[float, float]:
    """
    Returns the ratio between the contour and the smallest enclosing circle.

    :param contour: The contour to be compared
    :return: the ratio, the radius of the enclosing circle
    :rtype: float
    """
    _, enclosing_radius = cv2.minEnclosingCircle(contour)
    enclosing_circle_area = circle_area(enclosing_radius)
    contour_area = cv2.contourArea(contour)
    return float(contour_area) / enclosing_circle_area, enclosing_radius



from typing import Any, Union
from threading import Thread
import numpy as np
import cv2


class Camera:
    __slots__ = ("stream", "grabbed", "frame", "stopped", "camera_thread", "start_immediately")

    def __init__(self, source: Union[str, int, cv2.VideoCapture] = 0,
                 image_width: int = 320, image_height: int = 240, start_immediately=True):
        """
        Camera connects to and opens a connected camera and on constantly reads image from the camera.
        ovl.Camera is more real-time oriented and operates at a faster rate than opencv's VideoCapture, but is not
        suited for non-infinite streams

        The connected camera can be opened by using a camera number (index), url of ip camera, device file (/dev/video0)

        Note: as a result of increasing fps, Camera is not fit for opening video files or any other
        definite frame form of video.

        :param source: The source of the camera, can be a number, a device name or any other valid source
        for the cv2.VideoCapture object.
        :param image_width: The width of images to be captured in pixels
        :param image_height: The height of images to be captured in pixels
        :param start_immediately: The Camera has an inner thread that reads images, this determines if
        it should start immediately or be started by Camera.start manually.
        """

        self.stream = cv2.VideoCapture(source)

        if image_width:
            self.stream.set(cv2.CAP_PROP_FRAME_WIDTH, image_width)
        if image_height:
            self.stream.set(cv2.CAP_PROP_FRAME_HEIGHT, image_height)
        self.grabbed, self.frame = self.stream.read()
        self.stopped = False
        self.camera_thread: Union[None, Thread] = None
        self.start_immediately = start_immediately
        if start_immediately:
            self.start()

    def set_exposure(self, exposure_value: float) -> bool:
        """
        Sets the exposure for the camera.
        Value range depend on the camera some use the value as the exponential negative values or actual number of ms

        :param exposure_value: the exposure value to be set
        :return: if it was successful in setting the value
        """
        return self.stream.set(15, exposure_value)

    def start(self) -> "Camera":
        """
        Starts the image taking thread

        :return: self
        """
        if self.stopped:
            return self
        self.stopped = False
        self.camera_thread = Thread(target=self._update, args=()).start()
        return self

    def _update(self) -> None:
        """
        Takes a new image while not stopped

        """
        while not self.stopped:
            self.grabbed, self.frame = self.stream.read()

    def read(self) -> [bool, np.ndarray]:
        """
        Returns the return value (if getting the frame was successful and the frame itself

        This function is mainly for compatibility with cv2.VideoCapture, Camera.get_image() is
        recommended for common use.

        :return: if the image was taken successfully, the image
        :rtype: bool, numpy.array
        """
        return self.grabbed, self.frame

    def get_image(self) -> np.ndarray:
        """
        Returns the current image

        :return: numpy array of the image
        """
        return self.frame

    def stop(self) -> None:
        """
        Stops the camera thread.

        """
        self.stopped = True

    def release(self) -> None:
        """
        Stops the camera thread and releases the camera

        :return: None
        """
        self.stop()
        self.stream.release()

    def set(self, property_id: int, value: Union[float, bool, str, Any]) -> None:
        """
        Sets the value of the given property to the given value

        Properties are characteristics of the camera that can be changed to alter how images
        are returns, things from image width and height, contrast and saturation exposure etc.

        Not all cameras support all property changes, make sure the one you are using does.

        Read more at:
        https://docs.opencv.org/3.1.0/d8/dfe/classcv_1_1VideoCapture.html#aeb1644641842e6b104f244f049648f94

        import cv2 and use cv2.CAP_PROP_<PROPERTY NAME> or simply use the property id number (number depends
        on the property's location f.e CAP_PROP_FRAME_WIDTH is 3

        :param property_id: the property id (number)
        :param value: the value to be set, a number
        :return: None
        """
        self.stream.set(property_id, value)

    def get(self, property_id) -> Any:
        """
        Retrieves the value of a property based on its id

        :param property_id: the property number, for more information:
                        https://docs.opencv.org/3.1.0/d8/dfe/classcv_1_1VideoCapture.html
        :return: the value of the property
        """
        if isinstance(property_id, int):
            raise ValueError("Given Property id {} was not valid, please refer to the documentation: "
                             "https://docs.opencv.org/3.1.0/d8/dfe/classcv_1_1VideoCapture.html".format(property_id))
        if not 0 <= type(property_id) < 19:
            raise ValueError("Invalid property id ({}), please refer to the documentation".format(property_id))
        return self.stream.get(property_id)

    def is_opened(self) -> bool:
        """
        Returns true_shape if the camera is opened, false otherwise

        :return: if the camera is open
        """
        return self.stream.isOpened()

    def get_backend_name(self) -> str:
        """
        Gets the backend name of the camera using cv2.VideoCapture.getBackendName()

        :return: returns the backend name of a camera
        """
        return self.stream.getBackendName()



from typing import Union, Tuple, Any

import numpy as np

VisionLike = Union["Vision", "AmbientVision"]


Target = Union[np.ndarray, Tuple, Any]


def RangedNumber(start, end):
    class Hint:
        def __class_getitem__(cls, item):
            cls.start = item[0]
            cls.end = item[1]
            return cls
    return Union[float, Hint[start, end]]





from typing import Union

import cv2
import enum

import numpy as np

from .threshold import Threshold


class BinaryThresholdType(enum.IntEnum):
    Binary = cv2.THRESH_BINARY
    BinaryInverted = cv2.THRESH_BINARY_INV
    Truncate = cv2.THRESH_TRUNC
    ToZero = cv2.THRESH_TOZERO
    ToZeroInverted = cv2.THRESH_TOZERO_INV


def is_valid_threshold_type(threshold_type):
    for current_threshold_type in BinaryThresholdType.__members__.values():
        current_type_value = current_threshold_type.value
        if threshold_type == current_type_value or threshold_type == current_type_value + cv2.THRESH_OTSU:
            return True
    return False


class BinaryThreshold(Threshold):
    """
    Creates binary image (masks) from greyscale images (black and white images)
    The basic Binary Thresholding turns all pixels that are greater or equal to the threshold parameter
    and smaller or equal to the parameter upper_bound

    For more information about the algorithms used:
    https://docs.opencv.org/trunk/d7/d4d/tutorial_py_thresholding.html
    """

    def __init__(self, threshold: int = None, upper_bound: int = None,
                 threshold_type: Union[BinaryThresholdType, int] = BinaryThresholdType.Binary,
                 otsu=False):
        """


        :param threshold: minimum value that is defined to pass the threshold
        :param upper_bound: maximum
        :param threshold_type: what to do with pixels are determined to pass the threshold,
        
        :param otsu: when true uses otsu binarization (Bi-modal thresholding),
        when using otsu binarization threshold and upper_bound are *ignored* and are not used
        """
        self.threshold_type = threshold_type
        self.threshold = threshold
        self.upper_bound = upper_bound
        self.otsu = otsu

    def convert(self, image: np.ndarray) -> np.ndarray:
        """
        Converts a greyscale image to a binary image using Binary thresholding.
        :param image: an opened image (numpy ndarray)
        :return:
        """
        return cv2.threshold(image, self.threshold, self.upper_bound, self.threshold_type)

    def validate(self, *args, **kwargs) -> bool:
        """
        Validates that the parameters for the Threshold are valid

        :param args:
        :param kwargs:
        :return:
        """
        return 0 <= self.threshold <= self.upper_bound <= 255 and is_valid_threshold_type(self.threshold_type)



import numpy as np

from .contours import contour_center, distance_between_points
from .geometry import slope, x_intersection, y_intersection


def image_center(image_dimensions):
    """
    Calculates the center pixels of a given image dimension

    :param image_dimensions: (width, height) tuple of the size of the image
    :return: the (x, y) center of the image
    """
    return tuple(map(lambda dimension: dimension / 2, image_dimensions))


def distance_from_frame(point, image_dimensions):
    """
    Calculates the distance of a given point from the frame of the image based on the vector from the center
    and the point

    :param point: point (x,y tuple) or contour (numpy array)
    :param image_dimensions: the size of the image, (width, height)
    :return: the distance
    """
    point = contour_center(point) if type(point) is np.ndarray else point
    center_of_image = image_center(image_dimensions)
    line_slope = slope(point, center_of_image)
    intercept = - line_slope * point[0] + point[1]
    xs = (image_dimensions[0] - 1, 0)
    ys = (image_dimensions[1] - 1, 0)
    y_distances = [distance_between_points(y_intersection(line_slope, intercept, y), point) for y in ys]
    x_distances = [distance_between_points(x_intersection(line_slope, intercept, x), point) for x in xs]
    return min(y_distances + x_distances)


def image_size(image: np.ndarray) -> int:
    """
    Calculates the size of the given image in pixels

    :param image: ndarray image in pixels
    :return: the image size
    """
    return image.shape[0] * image.shape[1]



from typing import Any, Union

from .connection import Connection
from ..helpers.team_number_to_ip import team_number_to_ip

NetworkTables = None


class NetworkTablesConnection(Connection):
    """
    Note: In Order to use NetworkTablesConnection you must have pynetworktables installed.
    (It is automatically installed when installing ovl)

    A connection to that uses NetworkTables (The FRC network protocol).

    NetworkTables are a group of Dictionaries (Hash tables) that are shared by all computers in the network.
    In the case of the FRC it is created by the RoboRIO and shared by other
    computers on the network like: the driver station, any connected co-processor (like a raspberry pi) etc.

    NetworkTablesConnection creates defaults to writing to the Vision table.
    You can then read from /vision/vision_result the result sent.

    For Additional information about NetworkTables
    please refer to:
    https://docs.wpilib.org/en/latest/docs/software/networktables/networktables-intro.html
    """

    def __init__(self, roborio: str, table_name: str = "Vision", table_key: str = "vision_result"):
        """
        :param roborio: team number, roborio hostname or ip.
            for a team number 1234 valid inputs are:
              - 1234
              - roboRIO-1234-FRC.local
              - 10.12.34.2 (10.TE.AM.2)
              - a dynamic ip of the roborio

        NOTE: When using static ip (relevant inputs: 1234 and 10.12.34.2)
              make sure both your computer and the roborio are set to have a static ip
              for more information:
              https://docs.wpilib.org/en/latest/docs/networking/networking-introduction/index.html

        :param table_name: the table to use as a default "Vision" is recommended
        :param table_key: the key in the table to use as a default
        """
        from networktables import NetworkTables
        if roborio.isnumeric():
            roborio = team_number_to_ip(roborio)
        self.roborio_ip = roborio
        self.connection = NetworkTables.initialize(roborio)
        self.table_key = table_key
        self.table = NetworkTables.getTable(table_name)
        self.table_cache = {table_name: self.table}
        self.table_name = table_name

    def get_table(self, table):
        """
        Fetches a table by name, if table exists in cache it returns
        the cached table instead.
        If table is none it returns the default table (set in the constructor)

        :param table: the name of the table examples: "usage", "SmartDashboard", "vision"
        if it doesnt exist it is created
        :return: the table
        """
        from networktables import NetworkTables
        if table is None:
            table = self.table
        elif table in self.table:
            table = self.table_cache[table]
        else:
            table_name = table
            table = NetworkTables.getTable(table)
            self.table_cache[table_name] = table
        return table

    def send(self, data, table_key: Union[str, Any] = None,
             table: Union[str, Any] = None, *args, **kwargs) -> None:
        """
        A function used to put a value in a given table and key in the connected
        NetworkTable.
        For more information about NetworkTables functionality
        please refer to:
        https://docs.wpilib.org/en/latest/docs/software/networktables/networktables-intro.html

        :param data: the data to send (post to the NetworkTables)
        :param table_key: the specific table to read from
        :param table: the table to receive from. Examples are SmartDashboard
                      or Usage. Use "Vision" if you are not sure)
        """
        table_key = table_key if table_key else self.table_key
        table = self.get_table(table)
        return table.putValue(table_key, data)

    def receive(self, table_key: str = None, table: str = None, default_value=None, *args, **kwargs) -> Any:
        """
        Gets a value from a specific key and table, can be used to read values shared by
        all computers in the network.

        Default table and table keys are the ones initialized.

        .. code-block:: python

            import ovl

            TEAM_NUMBER = "1937"
            connection = ovl.NetworkTablesConnection(TEAM_NUMBER)

            connection.send("hello", table_key="my_key")

            print(connection.receive(table_key="my_key"))
            # prints "hello"


        :param table_key: the specific table to read from, None
        :param table: the table to receive from (Use "Vision" if you are not sure)
        :param default_value: the value to return if the given table key does not exist
        """
        table = self.get_table(table)
        table_key = table_key or self.table_key
        return table.getValue(table_key, default_value)



import numpy as np
from typing import List, Tuple
import types

from . import vision
from ..connections import connection
from ..detectors.detector import Detector


class AmbientVision:
    """
    A mesh of 2 vision objects that extends one of the visions at a time.

    The visions are swapped every main_amount of update_vision calls.
    This makes the "ambient_vision" run in the background once every multiple frames.

    The example includes only general instructions, 2 vision objects need to be defined to detect something

    An example will be as follows:

    .. code-block:: python

        vision1 = Vision(....)

        vision2 = Vision(....)

        vision_controller = AmbientVision(main_vision=vision1, ambient_vision=vision2, main_amount=3)

        while True:
            image = vision_controller.get_image()

            contours, image = vision_controller.detect(image)

            vision_controller.update_vision()


    you can get the current vision object using vision_controller.current_vision


    """

    def __init__(self, main_vision: vision.Vision, ambient_vision: vision.Vision,
                 main_amount: int, start_ambient: bool = False):
        self.main_vision = main_vision
        self.ambient_vision = ambient_vision
        self.main_amount = main_amount
        self.is_ambient = start_ambient
        if start_ambient:
            self.counter = 0
            self.current_vision = main_vision
        else:
            self.counter = main_amount
            self.current_vision = ambient_vision

    @property
    def connection(self) -> connection.Connection:
        return self.current_vision.connection

    @property
    def detector(self) -> Detector:
        return self.current_vision.detector

    @property
    def image_filters(self) -> List[types.FunctionType]:
        return self.current_vision.image_filters

    @property
    def directions(self):
        return self.current_vision.director

    def get_image(self) -> np.ndarray:
        """
        Take a picture using the current vision

        See Vision.get_image for more information
        """
        return self.current_vision.get_image()

    def apply_image_filters(self, image: np.ndarray) -> np.ndarray:
        """
        Applies all of the image filter of the current vision on the given image

        See Vision.apply_image_filters for more information
        """
        return self.current_vision.apply_image_filters(image)

    def detect(self, image: np.ndarray, verbose=False, *args, **kwargs) -> Tuple[List[np.ndarray], np.ndarray]:
        """
        Gets contours and applies all filters and returns the result,
        thus detecting the object according to the specifications in the vision,
        Uses the current vision

        :param image: image in which the vision should detect an object
        :param verbose: if true prints additional information about contour filtering
        :return: contours, ratio list (from the filter functions) and the filtered image
        """
        return self.current_vision.detect(image, verbose=verbose, *args, **kwargs)

    def send(self, data, *args, **kwargs):
        """
        Sends a given value to the current vision's connection object

        :param data: the data to be sent
        :param args: any other parameters to the specific connection object
        :param kwargs:
        :return: Whatever the underlying connection object returns, usually the parameter data
        """
        return self.current_vision.send(data, *args, **kwargs)

    def get_directions(self, contours, image, sorter=None):
        """
        Returns the directions for the given image and contours using the current vision

        :param contours: the final contours detected.
        :param image: the image where the contours were detected in.
        :param sorter: a sorter function that can be added to be applied before getting directions
        :return: the directions
        """
        return self.current_vision.get_directions(contours, image, sorter)

    def update_vision(self):
        """
        Increases the inner counter and swaps the ambient and the main vision
        after the set number of updates (self.main_amount)

        This is used to switch between the main vision and ambient vision
        """
        if self.counter < self.main_amount:
            self.counter += 1
            self.current_vision = self.main_vision
            self.is_ambient = False
        else:
            self.counter = 0
            self.current_vision = self.ambient_vision
            self.is_ambient = True
